﻿using System.Collections.Generic;
using System.Linq;

namespace NMag.Core
{
    public static class CurveValidator
    {
        public enum Axis
        {
            X = 0,
            Y = 1
        }

        public enum BoundType
        {
            Free,
            Limit,
            Include
        }

        public static Result AreValuesIncreasing(Curve curve, Axis axis, bool allowDoubles = false)
        {
            List<double> values = curve.Values[(int)axis];

            if(values.Count == 0)
            {
                return new Result(true);
            }

            double lastValue = values.First();
            int repeats = 0;

            for(int v = 1; v < values.Count; v++)
            {
                if (values[v] > lastValue)
                {
                    lastValue = values[v];
                    repeats = 0;
                    continue;
                }

                if (values[v] == lastValue)
                {
                    repeats++;

                    if(!allowDoubles)
                    {
                        return new Result(false, $"{axis} value ({values[v]}) is repeating at position {v + 1}.");
                    }

                    if (repeats > 1)
                    {
                        return new Result(false, $"{axis} value ({values[v]}) is repeating more than once at position {v + 1}.");
                    }

                    continue;
                }

                return new Result(false, $"{axis} value ({values[v]}) at position {v + 1} is lower than the previous value ({values[v - 1]}).");
            }

            return new Result(true);
        }

        public static Result AreValuesBetweenBounds(Curve curve, Axis axis, BoundType boundMin, BoundType boundMax, double min = 0, double max = 0)
        {
            if(curve.Points.Count == 0)
            {
                return new Result(true);
            }

            switch (boundMin)
            {
                case BoundType.Include:

                    if (curve.Values[(int)axis].Min() != min)
                    {
                        return new Result(false, $"Lowest {axis} value must be ({min}).");
                    }
                    break;

                case BoundType.Limit:

                    if (curve.Values[(int)axis].Min() < min)
                    {
                        return new Result(false, $"Lowest {axis} value ({curve.Values[(int)axis].Min()}) is lower than the allowed minimum ({min}).");
                    }
                    break;

                default:
                    break;
            }

            switch (boundMax)
            {
                case BoundType.Include:

                    if (curve.Values[(int)axis].Max() != max)
                    {
                        return new Result(false, $"Highest {axis} value must be ({max}).");
                    }
                    break;

                case BoundType.Limit:

                    if (curve.Values[(int)axis].Max() > max)
                    {
                        return new Result(false, $"Highest {axis} value ({curve.Values[(int)axis].Max()}) is higher than the allowed maximum ({max}).");
                    }
                    break;

                default:
                    break;
            }

            return new Result(true);
        }

        public static Result CheckXSteps(Curve curve, double stepSize = 0, double stepCount = 0)
        {
            List<double> values = curve.Values[0];

            if(stepCount > 0 && curve.Points.Count != stepCount)
            {
                return new Result(false, $"Curve must have {stepCount} points.");
            }

            if(curve.Points.Count < 2)
            {
                return new Result(true);
            }

            if(stepSize == 0)
            {
                stepSize = values[1] - values[0];
            }

            for (int v = 1; v < values.Count; v++)
            {
                if(values[v] - values[v - 1] != stepSize)
                {
                    return new Result(false, $"Difference between values {v} ({values[v - 1]}) and {v + 1} ({values[v]}) must be {stepSize}.");
                }
            }

            return new Result(true);
        }

        public static Result CheckUnitSum(Curve curve, double total)
        {
            List<Point> points = curve.Points;

            if (points.Count == 0)
            {
                return new Result(false, $"The sum of the area under the graph must be {total}.");
            }

            if(points.Count == 1)
            {
                if(points.First().X != total)
                {
                    return new Result(false, $"The sum of the area under the graph ({points.First().X}) must be {total}.");
                }
            }

            double sum = 0;
            double sumX = 0;

            for (int p = 1; p < points.Count; p++)
            {
                double dx = points[p].X - points[p - 1].X;
                double y = (points[p].Y + points[p - 1].Y) / 2;

                sumX += dx;
                sum += y * dx;
            }

            if(sum != total)
            {
                return new Result(false, $"The sum of the area under the graph ({sum}) must be {total}.");
            }

            return new Result(true);
        }
    }
}
