﻿using NMag.Core.Utilization;
using System.Collections.Generic;
using System.Linq;

namespace NMag.Core
{
    public enum CurveType
    {
        Routing,
        Hydro,
        Volume,
        SpillFlow,
        GateDischarge,
        Evaporation,
        LoadEfficiency,
        Peaking,
        HQCapacity,
        Preference,
        DailyLoad,
        VariableLoad,
        MinimumFlow,
        BypassRelease,
        MinimumLevel,
        MaximumLevel,
        Guide,
        RationingRule,
        SurplusRule,
        ReservoirRelease,
        ReservoirGuide,
        Storage,
        WaterLevel,
        Production
    }

    public class Curve
    {
        private string _code;

        public Curve(CurveType type, string code, int maxLength, Axis xAxis, Axis yAxis, List<Point> points)
        {
            Type = type;
            _code = code;
            MaxLength = maxLength;
            Points = points ?? new List<Point>();

            Axes = new Axis[] { xAxis, yAxis };
        }

        protected Curve(Curve curve)
        {
            if(curve == null) { return; }

            _code = curve._code;
            Type = curve.Type;

            Axes = curve.Axes.Select(a => a.Clone()).ToArray();
            MaxLength = curve.MaxLength;

            Points = curve.Points.Select(p => p.Clone()).ToList();

            Validation = curve.Validation;

            CanTranspose = curve.CanTranspose;
            UnitSum = curve.UnitSum;
            HasRepeatingX = curve.HasRepeatingX;

            FirstYear = curve.FirstYear;
            LastYear = curve.LastYear;

            DataSource = curve.DataSource;
        }

        public int Code
        {
            get
            {
                int.TryParse(_code.DigitsOnly(), out int code);
                return code;
            }
        }
        public CurveType Type { get; private set; }

        public Axis[] Axes { get; private set; } = new Axis[] { new Axis("", "", false), new Axis("", "", false) };
        public int MaxLength { get; private set; } = 0;

        public List<Point> Points { get; set; } = new List<Point>();
        public List<double>[] Values
        {
            get
            {
                List<double>[] values = new List<double>[] { new List<double>(), new List<double>() };

                Points.ForEach(p => 
                {
                    values[0].Add(p.X);
                    values[1].Add(p.Y);
                });

                return values;
            }
        }

        public delegate Result ValidationDelegate(Curve curve);
        public ValidationDelegate Validation;

        public bool CanTranspose { get; protected set; } = false;
        public double UnitSum { get; protected set; } = 0;
        public bool HasRepeatingX { get; protected set; } = false;
        public bool IsTimeSeries
        {
            get
            {
                return Axes[0].TimeScale != TimeScale.None;
            }
        }

        public int FirstYear { get; set; } = 0;
        public int LastYear { get; set; } = 0;

        public string DataSource { get; set; } = "";

        public Curve Clone()
        {
            return new Curve(this);
        }

        public void Normalize()
        {
            if(UnitSum == 0 || Points.Count < 2)
            {
                return;
            }

            double start = Points.First().X;
            double end = Points.Last().X;

            List<double> yValues = new List<double>();

            for(double x = start; x <= end; x++)
            {
                yValues.Add(Points.Interpolate(x));
            }

            double normalizeFactor = UnitSum / yValues.Sum();

            Points.ForEach(p => p.Y *= normalizeFactor);
        }

        public void ConvertToContinuous()
        {
            if(Points.Where(p => p.X == 0.0).Count() == 0)
            {
                return;
            }

            int surplusEnd = Points.Select(p => p.X).ToList().IndexOf(0.0);
            int rationingStart = Points.Select(p => p.X).ToList().LastIndexOf(0.0);

            double sum = 0;

            for(int i = surplusEnd; i >= 0; i--)
            {
                Points[i].X = sum += Points[i].X;
            }

            sum = 0;

            for (int i = rationingStart; i < Points.Count; i++)
            {
                Points[i].X = sum += Points[i].X;
            }
        }

        public void ConvertToDiscrete()
        {
            if (Points.Where(p => p.X == 0.0).Count() == 0)
            {
                return;
            }

            int surplusEnd = Points.Select(p => p.X).ToList().IndexOf(0.0);
            int rationingStart = Points.Select(p => p.X).ToList().LastIndexOf(0.0);

            double sum = 0;

            for (int i = surplusEnd; i >= 0; i--)
            {
                Points[i].X = Points[i].X - sum;
                sum += Points[i].X;
            }

            sum = 0;

            for (int i = rationingStart; i < Points.Count; i++)
            {
                Points[i].X = Points[i].X - sum;
                sum += Points[i].X;
            }
        }

        public Result Validate()
        {
            return Points.Count == 0 ?
                new Result(true) :
                Validation.Invoke(this) ?? 
                new Result(true);
        }
    }

    // General

    public class RoutingCurve : Curve
    {
        public RoutingCurve(RoutingCurve curve) : base(curve) { }

        public RoutingCurve(List<Point> points) 
            : base(CurveType.Routing, "600", 49, new Axis("Delay", "time step", false), new Axis("Discharge", "%", true), points)
        {
            UnitSum = 100;

            Validation = IsValid;
        }

        public Result IsValid(Curve curve)
        {
            return new List<Result>()
            {
                CurveValidator.AreValuesIncreasing(curve, CurveValidator.Axis.X),
                CurveValidator.AreValuesBetweenBounds(curve, CurveValidator.Axis.X, CurveValidator.BoundType.Include, CurveValidator.BoundType.Free, 1),
                CurveValidator.AreValuesBetweenBounds(curve, CurveValidator.Axis.Y, CurveValidator.BoundType.Limit, CurveValidator.BoundType.Limit, 0, 100),
                CurveValidator.CheckXSteps(curve, 1),
                CurveValidator.CheckUnitSum(curve, UnitSum)
            }.FirstOrDefault(r => !r.Succes);
        }
    }

    public class HydroCurve : Curve
    {
        public HydroCurve(HydroCurve curve) : base(curve) { }

        public HydroCurve(List<Point> points, int firstYear)
            : base(CurveType.Hydro, "0", 0, new Axis("Period", "day", false, TimeScale.Dates), new Axis("Discharge", "m3 s-1", true), points)
        {
            FirstYear = firstYear;

            Validation = IsValid;
        }

        public Result IsValid(Curve curve)
        {
            return new List<Result>()
            {
                CurveValidator.AreValuesIncreasing(curve, CurveValidator.Axis.X),
                CurveValidator.AreValuesBetweenBounds(curve, CurveValidator.Axis.Y, CurveValidator.BoundType.Limit, CurveValidator.BoundType.Free, 0),
                CurveValidator.CheckXSteps(curve, 1)
            }.FirstOrDefault(r => !r.Succes);
        }

        public HydroCurve GetAverage()
        {
            HydroCurve curve = new HydroCurve(this);
            curve.Points.Clear();

            Dictionary<int, double> dailySums = new Dictionary<int, double>();
            Dictionary<int, int> dayCounts = new Dictionary<int, int>();

            foreach(Point point in Points)
            {
                int day = (int)(point.X % 365);

                if(!dailySums.ContainsKey(day))
                {
                    dailySums.Add(day, 0);
                    dayCounts.Add(day, 0);
                }

                dailySums[day] += point.Y;
                dayCounts[day]++;
            }

            if(dailySums.Count == 0)
            {
                return curve;
            }

            for (int d = dailySums.Keys.Min(); d <= dailySums.Keys.Max(); d++)
            {
                curve.Points.Add(new Point(d, dailySums[d] / dayCounts[d]));
            }

            return curve;
        }
    }

    // Reservoir

    public class VolumeCurve : Curve
    {
        public VolumeCurve(VolumeCurve curve) : base(curve) { }

        public VolumeCurve(List<Point> points) 
            : base(CurveType.Volume, "200", 49, new Axis("Water level", "masl", true), new Axis("Volume", "mill m3", true), points)
        {
            CanTranspose = true;

            Validation = IsValid;
        }

        public Result IsValid(Curve curve)
        {
            return new List<Result>()
            {
                CurveValidator.AreValuesIncreasing(curve, CurveValidator.Axis.X),
                CurveValidator.AreValuesIncreasing(curve, CurveValidator.Axis.Y),
                CurveValidator.AreValuesBetweenBounds(curve, CurveValidator.Axis.Y, CurveValidator.BoundType.Limit, CurveValidator.BoundType.Free, 0)
            }.FirstOrDefault(r => !r.Succes);
        }
    }

    public class SpillFlowCurve : Curve
    {
        public SpillFlowCurve(SpillFlowCurve curve) : base(curve) { }

        public SpillFlowCurve(List<Point> points) 
            : base(CurveType.SpillFlow, "300", 49, new Axis("Water level", "m", true), new Axis("Discharge", "m3 s-1", true), points)
        {
            CanTranspose = true;

            Validation = IsValid;
        }

        public Result IsValid(Curve curve)
        {
            return new List<Result>()
            {
                CurveValidator.AreValuesIncreasing(curve, CurveValidator.Axis.X),
                CurveValidator.AreValuesIncreasing(curve, CurveValidator.Axis.Y),
                CurveValidator.AreValuesBetweenBounds(curve, CurveValidator.Axis.X, CurveValidator.BoundType.Include, CurveValidator.BoundType.Free, 0),
                CurveValidator.AreValuesBetweenBounds(curve, CurveValidator.Axis.Y, CurveValidator.BoundType.Limit, CurveValidator.BoundType.Free, 0)
            }.FirstOrDefault(r => !r.Succes);
        }
    }

    public class GateDischargeCurve : Curve
    {
        public GateDischargeCurve(GateDischargeCurve curve) : base(curve) { }

        public GateDischargeCurve(List<Point> points) 
            : base(CurveType.GateDischarge, "400", 49, new Axis("Head", "m", true), new Axis("Discharge", "m3 s-1", true), points)
        {
            CanTranspose = true;

            Validation = IsValid;
        }

        public Result IsValid(Curve curve)
        {
            return new List<Result>()
            {
                CurveValidator.AreValuesIncreasing(curve, CurveValidator.Axis.X),
                CurveValidator.AreValuesIncreasing(curve, CurveValidator.Axis.Y),
                CurveValidator.AreValuesBetweenBounds(curve, CurveValidator.Axis.X, CurveValidator.BoundType.Include, CurveValidator.BoundType.Free, 0),
                CurveValidator.AreValuesBetweenBounds(curve, CurveValidator.Axis.Y, CurveValidator.BoundType.Limit, CurveValidator.BoundType.Free, 0)
            }.FirstOrDefault(r => !r.Succes);
        }
    }

    public class EvaporationCurve : Curve
    {
        public EvaporationCurve(EvaporationCurve curve) : base(curve) { }

        // TODO fix time scale
        public EvaporationCurve(List<Point> points) 
            : base(CurveType.Evaporation, "500", 12, new Axis("Period", "month", false), new Axis("EP", "mm day-1", true), points)
        {
            Validation = IsValid;
        }

        public Result IsValid(Curve curve)
        {
            return new List<Result>()
            {
                CurveValidator.AreValuesIncreasing(curve, CurveValidator.Axis.X),
                CurveValidator.AreValuesBetweenBounds(curve, CurveValidator.Axis.X, CurveValidator.BoundType.Include, CurveValidator.BoundType.Include, 1, 12),
                CurveValidator.AreValuesBetweenBounds(curve, CurveValidator.Axis.Y, CurveValidator.BoundType.Limit, CurveValidator.BoundType.Free, 0),
                CurveValidator.CheckXSteps(curve, 1, 12)
            }.FirstOrDefault(r => !r.Succes);
        }
    }

    // Power Plant

    public class LoadEfficiencyCurve : Curve
    {
        public LoadEfficiencyCurve(LoadEfficiencyCurve curve) : base(curve) { }

        public LoadEfficiencyCurve(List<Point> points) 
            : base(CurveType.LoadEfficiency, "200", 49, new Axis("Load", "%", true), new Axis("Efficiency", "%", true), points)
        {
            Validation = IsValid;
        }

        public Result IsValid(Curve curve)
        {
            return new List<Result>()
            {
                CurveValidator.AreValuesIncreasing(curve, CurveValidator.Axis.X),
                CurveValidator.AreValuesBetweenBounds(curve, CurveValidator.Axis.X, CurveValidator.BoundType.Limit, CurveValidator.BoundType.Limit, 0, 100),
                CurveValidator.AreValuesBetweenBounds(curve, CurveValidator.Axis.Y, CurveValidator.BoundType.Limit, CurveValidator.BoundType.Limit, 0, 100)
            }.FirstOrDefault(r => !r.Succes);
        }
    }

    public class PeakingCurve : Curve
    {
        public PeakingCurve(PeakingCurve curve) : base(curve) { }

        public PeakingCurve(List<Point> points) 
            : base(CurveType.Peaking, "400", 24, new Axis("Period", "hour", false, TimeScale.Hours), new Axis("Load", "%", true), points)
        {
            UnitSum = 100;
        }

        public Result IsValid(Curve curve)
        {
            return new List<Result>()
            {
                CurveValidator.AreValuesIncreasing(curve, CurveValidator.Axis.X),
                CurveValidator.AreValuesBetweenBounds(curve, CurveValidator.Axis.X, CurveValidator.BoundType.Include, CurveValidator.BoundType.Include, 1, 24),
                CurveValidator.AreValuesBetweenBounds(curve, CurveValidator.Axis.Y, CurveValidator.BoundType.Limit, CurveValidator.BoundType.Limit, 0, 100),
                CurveValidator.CheckUnitSum(curve, 100)
            }.FirstOrDefault(r => !r.Succes);
        }
    }

    // Transfer

    public class HQCapacityCurve : Curve
    {
        public HQCapacityCurve(HQCapacityCurve curve) : base(curve) { }

        public HQCapacityCurve(List<Point> points) 
            : base(CurveType.HQCapacity, "300", 49, new Axis("Head", "m", true), new Axis("Discharge", "m3 s-1", true), points)
        {
            CanTranspose = true;

            Validation = IsValid;
        }

        public Result IsValid(Curve curve)
        {
            return new List<Result>()
            {
                CurveValidator.AreValuesIncreasing(curve, CurveValidator.Axis.X),
                CurveValidator.AreValuesIncreasing(curve, CurveValidator.Axis.Y),
                CurveValidator.AreValuesBetweenBounds(curve, CurveValidator.Axis.Y, CurveValidator.BoundType.Limit, CurveValidator.BoundType.Free, 0),
                CurveValidator.AreValuesBetweenBounds(curve, CurveValidator.Axis.Y, CurveValidator.BoundType.Limit, CurveValidator.BoundType.Free, 0)
            }.FirstOrDefault(r => !r.Succes);
        }
    }

    // Energy Market

    public class PreferenceCurve : Curve
    {
        public PreferenceCurve(PreferenceCurve curve) : base(curve) { }

        public PreferenceCurve(List<Point> points) 
            : base(CurveType.Preference, "0", 25, new Axis("Firm power deficit", "%", true), new Axis("Price", "Currency", true), points)
        {
            HasRepeatingX = true;

            Validation = IsValid;
        }

        public Result IsValid(Curve curve)
        {
            return new List<Result>()
            {
                CurveValidator.AreValuesIncreasing(curve, CurveValidator.Axis.X, true),
                CurveValidator.AreValuesBetweenBounds(curve, CurveValidator.Axis.X, CurveValidator.BoundType.Free, CurveValidator.BoundType.Include, 0, 100)
            }.FirstOrDefault(r => !r.Succes);
        }
    }

    public class DailyLoadCurve : Curve
    {
        public DailyLoadCurve(DailyLoadCurve curve) : base(curve) { }

        public DailyLoadCurve(List<Point> points) 
            : base(CurveType.DailyLoad, "1", 365, new Axis("Period", "day", false, TimeScale.Days), new Axis("Amount of annual", "%", true), points)
        {
            UnitSum = 100;

            Validation = IsValid;
        }

        public Result IsValid(Curve curve)
        {
            return new List<Result>()
            {
                CurveValidator.AreValuesIncreasing(curve, CurveValidator.Axis.X, true),
                CurveValidator.AreValuesBetweenBounds(curve, CurveValidator.Axis.X, CurveValidator.BoundType.Include, CurveValidator.BoundType.Include, 1, 365),
                CurveValidator.AreValuesBetweenBounds(curve, CurveValidator.Axis.Y, CurveValidator.BoundType.Limit, CurveValidator.BoundType.Limit, 0, 100),
                CurveValidator.CheckUnitSum(curve, 100)
            }.FirstOrDefault(r => !r.Succes);
        }
    }

    public class VariableLoadCurve : Curve
    {
        public VariableLoadCurve(VariableLoadCurve curve) : base(curve) { }

        public VariableLoadCurve(List<Point> points) 
            : base(CurveType.VariableLoad, "3", 40, new Axis("Period", "day", false, TimeScale.Days), new Axis("Amount of annual", "%", true), points)
        {
            UnitSum = 100;

            Validation = IsValid;
        }

        public Result IsValid(Curve curve)
        {
            return new List<Result>()
            {
                CurveValidator.AreValuesIncreasing(curve, CurveValidator.Axis.X, true),
                CurveValidator.AreValuesBetweenBounds(curve, CurveValidator.Axis.X, CurveValidator.BoundType.Include, CurveValidator.BoundType.Include, 1, 365),
                CurveValidator.AreValuesBetweenBounds(curve, CurveValidator.Axis.Y, CurveValidator.BoundType.Limit, CurveValidator.BoundType.Limit, 0, 100),
                CurveValidator.CheckUnitSum(curve, 100)
            }.FirstOrDefault(r => !r.Succes);
        }
    }

    // Restrictions

    public class MinimumFlowCurve : Curve
    {
        public MinimumFlowCurve(MinimumFlowCurve curve) : base(curve) { }

        public MinimumFlowCurve(List<Point> points) 
            : base(CurveType.MinimumFlow, "100", 49, new Axis("Period", "day", false, TimeScale.Days), new Axis("Discharge", "m3 s-1", true), points)
        {
            Validation = IsValid;   
        }

        public Result IsValid(Curve curve)
        {
            return new List<Result>()
            {
                CurveValidator.AreValuesIncreasing(curve, CurveValidator.Axis.X, true),
                CurveValidator.AreValuesBetweenBounds(curve, CurveValidator.Axis.X, CurveValidator.BoundType.Include, CurveValidator.BoundType.Include, 1, 365),
                CurveValidator.AreValuesBetweenBounds(curve, CurveValidator.Axis.Y, CurveValidator.BoundType.Limit, CurveValidator.BoundType.Free, 0, 0)
            }.FirstOrDefault(r => !r.Succes);
        }
    }

    public class BypassReleaseCurve : Curve
    {
        public BypassReleaseCurve(BypassReleaseCurve curve) : base(curve) { }

        public BypassReleaseCurve(List<Point> points) 
            : base(CurveType.BypassRelease, "200", 49, new Axis("Period", "day", false, TimeScale.Days), new Axis("Discharge", "m3 s-1", true), points)
        {
            Validation = IsValid;
        }

        public Result IsValid(Curve curve)
        {
            return new List<Result>()
            {
                CurveValidator.AreValuesIncreasing(curve, CurveValidator.Axis.X, true),
                CurveValidator.AreValuesBetweenBounds(curve, CurveValidator.Axis.X, CurveValidator.BoundType.Include, CurveValidator.BoundType.Include, 1, 365),
                CurveValidator.AreValuesBetweenBounds(curve, CurveValidator.Axis.Y, CurveValidator.BoundType.Limit, CurveValidator.BoundType.Free, 0, 0)
            }.FirstOrDefault(r => !r.Succes);
        }
    }

    public class MinimumLevelCurve : Curve
    {
        public MinimumLevelCurve(MinimumLevelCurve curve) : base(curve) { }

        public MinimumLevelCurve(List<Point> points) 
            : base(CurveType.MinimumLevel, "300", 49, new Axis("Period", "day", false, TimeScale.Days), new Axis("Reservoir volume", "%", true), points)
        {
            Validation = IsValid;
        }

        public Result IsValid(Curve curve)
        {
            return new List<Result>()
            {
                CurveValidator.AreValuesIncreasing(curve, CurveValidator.Axis.X, true),
                CurveValidator.AreValuesBetweenBounds(curve, CurveValidator.Axis.X, CurveValidator.BoundType.Include, CurveValidator.BoundType.Include, 1, 365),
                CurveValidator.AreValuesBetweenBounds(curve, CurveValidator.Axis.Y, CurveValidator.BoundType.Limit, CurveValidator.BoundType.Limit, 0, 100)
            }.FirstOrDefault(r => !r.Succes);
        }
    }

    public class MaximumLevelCurve : Curve
    {
        public MaximumLevelCurve(MaximumLevelCurve curve) : base(curve) { }

        public MaximumLevelCurve(List<Point> points) 
            : base(CurveType.MaximumLevel, "400", 49, new Axis("Period", "day", false, TimeScale.Days), new Axis("Reservoir volume", "%", true), points)
        {
            Validation = IsValid;
        }

        public Result IsValid(Curve curve)
        {
            return new List<Result>()
            {
                CurveValidator.AreValuesIncreasing(curve, CurveValidator.Axis.X, true),
                CurveValidator.AreValuesBetweenBounds(curve, CurveValidator.Axis.X, CurveValidator.BoundType.Include, CurveValidator.BoundType.Include, 1, 365),
                CurveValidator.AreValuesBetweenBounds(curve, CurveValidator.Axis.Y, CurveValidator.BoundType.Limit, CurveValidator.BoundType.Limit, 0, 100)
            }.FirstOrDefault(r => !r.Succes);
        }
    }

    // EnMag Operational Strategy
    
    public class GuideCurve : Curve
    {
        public GuideCurve(GuideCurve curve) : base(curve) { }

        public GuideCurve(List<Point> points) 
            : base(CurveType.Guide, "500", 49, new Axis("Period", "day", false, TimeScale.Days), new Axis("Release discharge", "m3 s-1", true), points)
        {
            Validation = IsValid;
        }

        public Result IsValid(Curve curve)
        {
            return new List<Result>()
            {
                CurveValidator.AreValuesIncreasing(curve, CurveValidator.Axis.X, true),
                CurveValidator.AreValuesBetweenBounds(curve, CurveValidator.Axis.X, CurveValidator.BoundType.Include, CurveValidator.BoundType.Include, 1, 365),
                CurveValidator.AreValuesBetweenBounds(curve, CurveValidator.Axis.Y, CurveValidator.BoundType.Limit, CurveValidator.BoundType.Free, 0)
            }.FirstOrDefault(r => !r.Succes);
        }
    }

    public class RationingRuleCurve : Curve
    {
        public RationingRuleCurve(RationingRuleCurve curve) : base(curve) { }

        public RationingRuleCurve(List<Point> points) 
            : base(CurveType.RationingRule, "600A", 49, new Axis("Period", "day", false, TimeScale.Days), new Axis("Reservoir volume", "%", true), points)  // TODO: Check if code is used correctly
        {
            Validation = IsValid;
        }

        public Result IsValid(Curve curve)
        {
            return new List<Result>()
            {
                CurveValidator.AreValuesIncreasing(curve, CurveValidator.Axis.X, true),
                CurveValidator.AreValuesBetweenBounds(curve, CurveValidator.Axis.X, CurveValidator.BoundType.Include, CurveValidator.BoundType.Include, 1, 365),
                CurveValidator.AreValuesBetweenBounds(curve, CurveValidator.Axis.Y, CurveValidator.BoundType.Limit, CurveValidator.BoundType.Limit, 0, 100)
            }.FirstOrDefault(r => !r.Succes);
        }
    }

    public class SurplusRuleCurve : Curve
    {
        public SurplusRuleCurve(SurplusRuleCurve curve) : base(curve) { }

        public SurplusRuleCurve(List<Point> points) 
            : base(CurveType.SurplusRule, "600B", 49, new Axis("Period", "day", false, TimeScale.Days), new Axis("Reservoir volume", "%", true), points)
        {
            Validation = IsValid;
        }

        public Result IsValid(Curve curve)
        {
            return new List<Result>()
            {
                CurveValidator.AreValuesIncreasing(curve, CurveValidator.Axis.X, true),
                CurveValidator.AreValuesBetweenBounds(curve, CurveValidator.Axis.X, CurveValidator.BoundType.Include, CurveValidator.BoundType.Include, 1, 365),
                CurveValidator.AreValuesBetweenBounds(curve, CurveValidator.Axis.Y, CurveValidator.BoundType.Limit, CurveValidator.BoundType.Limit, 0, 100)
            }.FirstOrDefault(r => !r.Succes);
        }
    }

    // NMag Operational Strategy

    public class ReservoirReleaseCurve : Curve
    {
        public ReservoirReleaseCurve(ReservoirReleaseCurve curve) : base(curve) { }

        public ReservoirReleaseCurve(List<Point> points) 
            : base(CurveType.ReservoirRelease, "400", 49, new Axis("Period", "day", false, TimeScale.Days), new Axis("Release discharge", "% of annual", true), points)
        {
            UnitSum = 100;
        }

        public Result IsValid(Curve curve)
        {
            return new List<Result>()
            {
                CurveValidator.AreValuesIncreasing(curve, CurveValidator.Axis.X, true),
                CurveValidator.AreValuesBetweenBounds(curve, CurveValidator.Axis.X, CurveValidator.BoundType.Include, CurveValidator.BoundType.Include, 1, 365),
                CurveValidator.AreValuesBetweenBounds(curve, CurveValidator.Axis.Y, CurveValidator.BoundType.Limit, CurveValidator.BoundType.Limit, 0, 100),
                CurveValidator.CheckUnitSum(curve, UnitSum)
            }.FirstOrDefault(r => !r.Succes);
        }
    }

    public class ReservoirGuideCurve : Curve
    {
        public ReservoirGuideCurve(ReservoirGuideCurve curve) : base(curve) { }

        public ReservoirGuideCurve(List<Point> points) 
            : base(CurveType.ReservoirGuide, "500", 49, new Axis("Period", "day", false, TimeScale.Days), new Axis("Reservoir volume", "%", true), points)
        {
            Validation = IsValid;
        }

        public Result IsValid(Curve curve)
        {
            return new List<Result>()
            {
                CurveValidator.AreValuesIncreasing(curve, CurveValidator.Axis.X, true),
                CurveValidator.AreValuesBetweenBounds(curve, CurveValidator.Axis.X, CurveValidator.BoundType.Include, CurveValidator.BoundType.Include, 1, 365),
                CurveValidator.AreValuesBetweenBounds(curve, CurveValidator.Axis.Y, CurveValidator.BoundType.Limit, CurveValidator.BoundType.Limit, 0, 100)
            }.FirstOrDefault(r => !r.Succes);
        }
    }

    // Results

    public class StorageCurve : Curve
    {
        public StorageCurve(StorageCurve curve) : base(curve) { }

        public StorageCurve(List<Point> points, int firstYear)
            : base(CurveType.Storage, "0", 0, new Axis("Period", "day", false, TimeScale.Dates), new Axis("Volume", "mill m3", true), points)
        {
            FirstYear = firstYear;

            Validation = IsValid;
        }

        public Result IsValid(Curve curve)
        {
            return new List<Result>()
            {
                CurveValidator.AreValuesIncreasing(curve, CurveValidator.Axis.X),
                CurveValidator.AreValuesBetweenBounds(curve, CurveValidator.Axis.Y, CurveValidator.BoundType.Limit, CurveValidator.BoundType.Free, 0),
                CurveValidator.CheckXSteps(curve, 1)
            }.FirstOrDefault(r => !r.Succes);
        }

        public StorageCurve GetAverage()
        {
            StorageCurve curve = new StorageCurve(this);
            curve.Points.Clear();

            Dictionary<int, double> dailySums = new Dictionary<int, double>();
            Dictionary<int, int> dayCounts = new Dictionary<int, int>();

            foreach (Point point in Points)
            {
                int day = (int)(point.X % 365);

                if (!dailySums.ContainsKey(day))
                {
                    dailySums.Add(day, 0);
                    dayCounts.Add(day, 0);
                }

                dailySums[day] += point.Y;
                dayCounts[day]++;
            }

            if (dailySums.Count == 0)
            {
                return curve;
            }

            for (int d = dailySums.Keys.Min(); d <= dailySums.Keys.Max(); d++)
            {
                curve.Points.Add(new Point(d, dailySums[d] / dayCounts[d]));
            }

            return curve;
        }
    }

    public class WaterLevelCurve : Curve
    {
        public WaterLevelCurve(WaterLevelCurve curve) : base(curve) { }

        public WaterLevelCurve(List<Point> points, int firstYear)
            : base(CurveType.WaterLevel, "0", 0, new Axis("Period", "day", false, TimeScale.Dates), new Axis("Water Level", "masl", true), points)
        {
            FirstYear = firstYear;

            Validation = IsValid;
        }

        public Result IsValid(Curve curve)
        {
            return new List<Result>()
            {
                CurveValidator.AreValuesIncreasing(curve, CurveValidator.Axis.X),
                CurveValidator.AreValuesBetweenBounds(curve, CurveValidator.Axis.Y, CurveValidator.BoundType.Limit, CurveValidator.BoundType.Free, 0),
                CurveValidator.CheckXSteps(curve, 1)
            }.FirstOrDefault(r => !r.Succes);
        }

        public WaterLevelCurve GetAverage()
        {
            WaterLevelCurve curve = new WaterLevelCurve(this);
            curve.Points.Clear();

            Dictionary<int, double> dailySums = new Dictionary<int, double>();
            Dictionary<int, int> dayCounts = new Dictionary<int, int>();

            foreach (Point point in Points)
            {
                int day = (int)(point.X % 365);

                if (!dailySums.ContainsKey(day))
                {
                    dailySums.Add(day, 0);
                    dayCounts.Add(day, 0);
                }

                dailySums[day] += point.Y;
                dayCounts[day]++;
            }

            if (dailySums.Count == 0)
            {
                return curve;
            }

            for (int d = dailySums.Keys.Min(); d <= dailySums.Keys.Max(); d++)
            {
                curve.Points.Add(new Point(d, dailySums[d] / dayCounts[d]));
            }

            return curve;
        }
    }

    public class ProductionCurve : Curve
    {
        public ProductionCurve(ProductionCurve curve) : base(curve) { }

        public ProductionCurve(List<Point> points, int firstYear)
            : base(CurveType.Production, "0", 0, new Axis("Period", "day", false, TimeScale.Dates), new Axis("Average Production", "MW", true), points)
        {
            FirstYear = firstYear;

            Validation = IsValid;
        }

        public Result IsValid(Curve curve)
        {
            return new List<Result>()
            {
                CurveValidator.AreValuesIncreasing(curve, CurveValidator.Axis.X),
                CurveValidator.AreValuesBetweenBounds(curve, CurveValidator.Axis.Y, CurveValidator.BoundType.Limit, CurveValidator.BoundType.Free, 0),
                CurveValidator.CheckXSteps(curve, 1)
            }.FirstOrDefault(r => !r.Succes);
        }

        public ProductionCurve GetAverage()
        {
            ProductionCurve curve = new ProductionCurve(this);
            curve.Points.Clear();

            Dictionary<int, double> dailySums = new Dictionary<int, double>();
            Dictionary<int, int> dayCounts = new Dictionary<int, int>();

            foreach (Point point in Points)
            {
                int day = (int)(point.X % 365);

                if (!dailySums.ContainsKey(day))
                {
                    dailySums.Add(day, 0);
                    dayCounts.Add(day, 0);
                }

                dailySums[day] += point.Y;
                dayCounts[day]++;
            }

            if (dailySums.Count == 0)
            {
                return curve;
            }

            for (int d = dailySums.Keys.Min(); d <= dailySums.Keys.Max(); d++)
            {
                curve.Points.Add(new Point(d, dailySums[d] / dayCounts[d]));
            }

            return curve;
        }
    }
}