﻿using Newtonsoft.Json;
using System.Collections.Generic;
using System.Linq;

namespace NMag.Core
{
    public class Project
    {
        public Project(string name)
        {
            Name = name;
        }

        private Project(Project project)
        {
            Name = project.Name;
            ProjectSettings = project.ProjectSettings.Clone();
            Modules = project.Modules.Select(m => m.Clone()).ToList();
            EnergyMarket = project.EnergyMarket.Clone();

            EnMagMode = project.EnMagMode;

            Hydrology = project.Hydrology.Clone();
            JobControl = project.JobControl.Clone();
        }

        [JsonIgnore]
        public string FileName
        {
            get { return PathFinder.GetFullFileName(PathFinder.FileType.Project, Name); }
        }
        public string Name { get; set; }

        public ProjectSettings ProjectSettings = new ProjectSettings();

        public List<Module> Modules { get; private set; } = new List<Module>();

        public EnergyMarket EnergyMarket { get; set; } = new EnergyMarket();

        public bool EnMagMode { get; set; } = false;

        public Hydrology Hydrology { get; set; } = new Hydrology();
        public JobControl JobControl { get; set; } = new JobControl();       

        public Result CreateModules(List<Module> modules)
        {
            string report = "Invalid modules:";

            foreach(int number in modules.Select(m => m.Number).Distinct())
            {
                List<Module> moduleSet = modules.Where(m => m.Number == number).ToList();

                if(number < 0)
                {
                    report += string.Join("\n", modules.Select(m => $"{number} - {m.Number}"));
                    continue;
                }

                if(moduleSet.Count > 2)
                {
                    report += string.Join("\n", modules.Select(m => $"{number} - {m.Number}"));
                    continue;
                }

                if(moduleSet.Count == 2)
                {
                    ReservoirModule reservoir = ModuleHelper.GetReservoirs(moduleSet).FirstOrDefault();
                    PowerPlantModule powerPlant = ModuleHelper.GetPowerPlants(moduleSet).FirstOrDefault();

                    if (reservoir == null || powerPlant == null)
                    {
                        report += string.Join("\n", modules.Select(m => $"{number} - {m.Number}"));
                        continue;
                    }

                    Modules.Add(new CombinedModule(reservoir, powerPlant.PowerPlant));
                    continue;
                }

                Modules.Add(moduleSet.First());
            }

            List<Module> reservoirModules = modules.Where(m => m.ModuleType == ModuleType.Reservoir || m.ModuleType == ModuleType.Combined).ToList();

            if(reservoirModules.Count == 1 && reservoirModules.First().EnMagOperationalStrategy != null) { EnMagMode = true; }

            return new Result(true, report == "Invalid modules:" ? "" : report);
        }

        public Result Save(bool overwrite)
        {
            return ProjectHelper.SaveProject(this, overwrite);
        }

        public void SortModules()
        {
            Modules = Modules.OrderBy(m => m.Number).ToList();
        }

        public Project Clone()
        {
            return new Project(this);
        }

        public Result Validate()
        {
            Result result;

            if(Name == "") { return new Result(false, "No project name given."); }

            List<string> moduleErrors = Modules.Where(m => !m.Validate().Succes).Select(m => $"#{m.Number}" + m.Validate().Message).ToList();
            if(moduleErrors.Count > 0) { return new Result(false, "Invalid modules:\n" + string.Join("\n", moduleErrors)); }

            //result = EnergyMarket.Validate();
            //if (!result.Succes) { return result; }

            result = Hydrology.Validate();
            if (!result.Succes) { return result; }

            //result = JobControl.Validate();
            //if (!result.Succes) { return result; }

            return new Result(true);
        }
    }
}
