﻿using System;
using System.Collections.Generic;
using System.Linq;

namespace NMag.Core.FlowDiagram
{
    public class FlowDiagramFactory
    {
        private List<Module> _modules;

        private FlowDiagram _flowDiagram = new FlowDiagram();

        private List<FlowConnectionGroup> _flowConnections = new List<FlowConnectionGroup>();

        public FlowDiagram CreateFlowDiagram(List<Module> modules)
        {
            _modules = modules;

            CreateModuleConnections();

            CombineFlowChains();

            LinkFlowChains();

            SetFlowChainLevels();

            SetFlowChainColumns();

            ConnectModules();

            SetRowHeights();

            return _flowDiagram;
        }

        private void SetRowHeights()
        {
            int row = 0;
            List<Module> modules;

            do
            {
                modules = _modules.Where(m => GetRow(m.Number) == row).ToList();
                if (modules.Where(m => m.ModuleType == ModuleType.Combined).Count() > 0)
                {
                    _flowDiagram.HigherRows.Add(row);
                }
                row++;

            } while (modules.Count > 0);
        }

        private void ConnectModules()
        {
            foreach(Module module in _modules)
            {
                foreach(int destination in module.RoutingNumbers.Distinct().Where(n => n != 0))
                {
                    _flowDiagram.Connections.Add(
                        new FlowConnection(
                            module.Number, 
                            CreatePath(module.Number, destination), 
                            (FlowConnectionType)module.RoutingNumbers
                                .ToList()
                                .IndexOf(destination)));
                }
            }
        }

        private int GetColumn(int moduleNumber)
        {
            return _flowDiagram.FlowChains.First(c => c.ModuleNumbers.Contains(moduleNumber)).Column;
        }

        private int GetRow(int moduleNumber)
        {
            FlowChain chain = _flowDiagram.FlowChains.First(c => c.ModuleNumbers.Contains(moduleNumber));
            int row = GetLevelRow(chain.Level) + chain.ModuleNumbers.IndexOf(moduleNumber);

            _flowDiagram.HighestRow = Math.Max(row, _flowDiagram.HighestRow);
            return row;
        }

        private List<System.Windows.Point> CreatePath(int moduleStart, int moduleEnd)
        {
            int columnStart = GetColumn(moduleStart);
            int columnEnd = GetColumn(moduleEnd);
            int rowStart = GetRow(moduleStart);
            int rowEnd = GetRow(moduleEnd);

            List<System.Windows.Point> points = new List<System.Windows.Point>()
            {
                new System.Windows.Point(columnStart, rowStart),
                new System.Windows.Point(columnEnd, rowEnd)
            };

            if(columnStart != columnEnd)
            {
                int offset = Math.Max(Math.Sign(columnStart - columnEnd), 0);

                points.Insert(1, new System.Windows.Point(columnEnd + offset, rowStart));
                points.Insert(2, new System.Windows.Point(columnEnd + offset, rowEnd));
                return points;
            }

            if(!IsPathClear(new int[] { columnStart, columnEnd }, new int[] { rowStart, rowEnd }))
            {
                int offset = columnStart >= 0 ? 1 : 0;

                points.Insert(1, new System.Windows.Point(columnStart + offset, rowStart));
                points.Insert(2, new System.Windows.Point(columnStart + offset, rowEnd));
            }

            return points;
        }

        private bool IsPathClear(int[] columns, int[] rows)
        {
            if (columns[0] != columns[1]) { return true; }

            return _flowDiagram.FlowChains
                .Where(c => columns.Contains(c.Column))
                .Select(c => c.ModuleNumbers
                    .Where(n => GetRow(n) < rows[0] &&
                        GetRow(n) > rows[1])
                    .Count())
                .Where(c => c > 0)
                .Count() == 0;
        }

        private int GetLevelRow(int level)
        {
            return _flowDiagram.FlowLevelHeights.GetRange(0, level).Sum();
        }

        private void CreateModuleConnections()
        {
            foreach (Module module in _modules)
            {
                FindConnections(module.Number);
                _flowDiagram.FlowChains.Add(new FlowChain(module.Number));
            }
        }

        private void CombineFlowChains()
        {
            while (_flowDiagram.FlowChains.Count(c => !c.IsComplete && !c.IsAbsorbed) > 0)
            {
                FlowChain chain = _flowDiagram.FlowChains.FirstOrDefault(c => !c.IsComplete && !c.IsAbsorbed);

                FlowConnectionGroup connections = _flowConnections.FirstOrDefault(c => c.ModuleNumber == chain.Highest);

                if (connections.Sources.Count != 1)
                {
                    chain.IsComplete = true;
                    continue;
                }

                FlowConnectionGroup upstreamConnection = _flowConnections.FirstOrDefault(c => c.ModuleNumber == connections.Sources.FirstOrDefault());

                if (upstreamConnection.Sinks.Count == 1)
                {
                    FlowChain upstreamChain = _flowDiagram.FlowChains.FirstOrDefault(c => c.Lowest == upstreamConnection.ModuleNumber);

                    chain.ModuleNumbers.AddRange(upstreamChain.ModuleNumbers);
                    upstreamChain.IsAbsorbed = true;
                    continue;
                }

                chain.IsComplete = true;
            }

            _flowDiagram.FlowChains.RemoveAll(c => c.IsAbsorbed);
        }

        private void LinkFlowChains()
        {
            foreach (FlowChain chain in _flowDiagram.FlowChains)
            {
                _flowConnections.Where(c => c.Sinks.Contains(chain.Highest))
                    .ToList()
                    .ForEach(co => chain.Sources.Add(_flowDiagram.FlowChains
                        .FirstOrDefault(ch => ch.Lowest == co.ModuleNumber).Number));

                _flowConnections.Where(c => c.Sources.Contains(chain.Lowest))
                    .ToList()
                    .ForEach(co => chain.Sinks.Add(_flowDiagram.FlowChains
                        .FirstOrDefault(ch => ch.Highest == co.ModuleNumber).Number));
            }
        }

        private int GetColumn(int sinkColumn, int index, int total)
        {
            index += (1 - total % 2);
            return sinkColumn + (index % 2 - 1) * (index / 2) + index % 2;
        }

        private void ResolveIntersections()
        {
            bool intersectionFound = false;

            do
            {
                intersectionFound = false;

                foreach (FlowChain chain1 in _flowDiagram.FlowChains)
                {
                    foreach(FlowChain chain2 in _flowDiagram.FlowChains.Where(c => c != chain1))
                    {
                        if(chain1.Level == chain2.Level && chain1.Column == chain2.Column)
                        {
                            intersectionFound = true;
                            chain1.Column += (Math.Sign(chain1.Column) == 0) ? 1 : Math.Sign(chain1.Column);
                            break;
                        }
                    }
                    if (intersectionFound) { break; }
                }
            } while (intersectionFound);
        }

        private void SetFlowChainColumns()
        {
            for (int l = 0; l < _flowDiagram.FlowLevelHeights.Count; l++)
            {
                List<FlowChain> chains = _flowDiagram.FlowChains.Where(c => c.Level == l).ToList();

                List<FlowChain> releaseSinks = new List<FlowChain>();

                foreach(int sourceNumber in chains.Select(c => c.Lowest))
                {
                    Module sourceModule = _modules.First(m => m.Number == sourceNumber);
                    FlowChain sink = _flowDiagram.FlowChains.FirstOrDefault(c => c.ModuleNumbers.Contains(sourceModule.ReleaseModuleNumber));

                    if(sink != null) { releaseSinks.Add(sink); }
                }

                releaseSinks = releaseSinks.Distinct().ToList();

                foreach(FlowChain releaseSink in releaseSinks.Distinct())
                {
                    List<FlowChain> sourceChains = chains.Where(c => c.Sinks.Intersect(releaseSink?.ModuleNumbers ?? new List<int>() { 0}).Count() > 0 || c.Sinks.Count == 0).ToList();

                    for (int c = 0; c < sourceChains.Count; c++)
                    {
                        int sinkColumn = releaseSink == null ? 0 : 
                            releaseSink.Column;

                        chains[c].Column = GetColumn(sinkColumn, c, sourceChains.Count);
                    }
                }

                ResolveIntersections();
            }
        }

        private List<FlowChain> GetNextLevelFlowChains(List<int> passedSinks)
        {
            return _flowDiagram.FlowChains
                .Where(c => c.Level == -1 && passedSinks.Intersect(c.Sinks).Count() == c.Sinks.Count())
                .ToList();
        }

        private void SetFlowChainLevels()
        {
            List<int> passedSinks = new List<int>();
            int level = 0;

            while (passedSinks.Count < _flowDiagram.FlowChains.Count)
            {
                int height = 0;

                foreach (FlowChain chain in GetNextLevelFlowChains(passedSinks))
                {
                    height = Math.Max(height, chain.ModuleNumbers.Count);

                    chain.Level = level;
                    passedSinks.Add(chain.Number);
                }

                _flowDiagram.FlowLevelHeights.Add(height);

                level++;
            }
        }

        private void FindConnections(int moduleNumber)
        {
            Module center = _modules.FirstOrDefault(m => m.Number == moduleNumber);
            FlowConnectionGroup connections = _flowConnections.FirstOrDefault(c => c.ModuleNumber == moduleNumber);

            if (connections == null)
            {
                connections = new FlowConnectionGroup(center.Number);
                _flowConnections.Add(connections);
            }

            foreach (Module module in _modules.Where(m => m.Number != center.Number))
            {
                if (FlowsToModule(module, center) && !connections.Sources.Contains(module.Number))
                {
                    connections.Sources.Add(module.Number);
                    continue;
                }

                if (FlowsToModule(center, module) && !connections.Sinks.Contains(module.Number))
                {
                    connections.Sinks.Add(module.Number);
                }
            }
        }

        private bool FlowsToModule(Module source, Module sink)
        {
            if (source.RoutingNumbers.Contains(sink.Number))
            {
                return true;
            }

            return false;
        }

        // TODO Remove
        private List<string> FlowConnectionsToStrings()
        {
            List<string> output = new List<string>();

            foreach (FlowConnectionGroup connection in _flowConnections)
            {
                string upstream = string.Join(",", connection.Sources);
                string downstream = string.Join(",", connection.Sinks);

                string line = string.Format("{0} up: {1} down: {2}", connection.ModuleNumber, upstream, downstream);

                output.Add(line);
            }

            return output;
        }
        private List<string> FlowChainsToStrings()
        {
            List<string> output = new List<string>();

            List<FlowChain> chains = _flowDiagram.FlowChains.Where(c => !c.IsAbsorbed).ToList();

            for (int l = 0; l <= chains.Select(c => c.Level).Max(); l++)
            {
                output.Add(string.Format("Level {0}:", l.ToString()));

                foreach (FlowChain chain in chains.Where(c => c.Level == l))
                {
                    string chainUnits = string.Join(",", chain.ModuleNumbers);

                    string line = string.Format("Chain {0} units: {1} sources: {2} sinks: {3}", _flowDiagram.FlowChains.IndexOf(chain), 
                                                                                                chainUnits, 
                                                                                                string.Join(",", chain.Sources), 
                                                                                                string.Join(",", chain.Sinks));

                    output.Add(line);
                }
            }

            return output;
        }
    }
}