﻿using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Threading;

namespace NMag.Core.Simulation
{
    public enum ClearLevel
    {
        All,
        Minimum,
        None
    }

    public class Simulation
    {
        public Simulation(string name)
        {
            Name = name == "" ? "Simulation" : name;
        }

        public string Name { get; private set; }
        public SimulationResult Result { get; private set; }
        public List<string> Errors { get; private set; }

        public SimulationResult Run(Project project, ClearLevel clearLevel)
        {
            Result = null;
            Errors = new List<string>();

            if (project == null) { return Result; }

            project.JobControl.OutputInBinary = true;
            project.JobControl.BinaryFileName = "nMAGbin.txt";

            if (!PrepareSimulation(project, clearLevel)) { return Result; }

            ProcessStartInfo startInfo = new ProcessStartInfo()
            {
                CreateNoWindow = false,
                FileName = GetSimulatorPath(),
                WindowStyle = ProcessWindowStyle.Hidden,
                WorkingDirectory = Path.GetDirectoryName(GetSimulatorPath())
            };

            using (Process process = Process.Start(startInfo)) { process.WaitForExit(); }

            Thread.Sleep(100);

            if (File.Exists(GetBinaryOutputPath()))
            {
                List<Module> modules = project.Modules
                    .Where(m => m == project.Modules
                        .Where(n => n.Number == m.Number)
                        .First())
                    .Select(m => m.Clone())
                    .ToList();

                Result = new SimulationResult(modules, project);

                if(!new ResultReader().Read(Result, GetBinaryOutputPath()).Succes) { return Result = null; }
            }

            if(File.Exists(GetErrorPath()))
            {
                try
                {
                    Errors = File.ReadAllLines(GetErrorPath()).ToList();
                }
                catch { }

                if(Errors.Count(e => e.Replace(" ", "") != "") > 4)
                {
                    return Result = null;
                }
            }

            return Result;
        }

        private string GetSimulatorPath()
        {
            return Path.Combine(PathFinder.GetFolder(PathFinder.Folder.Simulation), "nmag_DOS.exe");
        }

        private string GetResultsPath()
        {
            return Path.Combine(PathFinder.GetFolder(PathFinder.Folder.Simulation), "nMAGres.txt");
        }

        private string GetBinaryOutputPath()
        {
            return Path.Combine(PathFinder.GetFolder(PathFinder.Folder.Simulation), "nMAGbin.txt");
        }

        private string GetErrorPath()
        {
            return Path.Combine(PathFinder.GetFolder(PathFinder.Folder.Simulation), "nMAGerr.txt");
        }

        private string GetRunoffSeriesPath(string name)
        {
            return Path.Combine(PathFinder.GetFolder(PathFinder.Folder.Simulation), name + ".vsf");
        }

        private void DeleteFile(string fileName)
        {
            try
            {
                File.Delete(fileName);
            }
            catch { }
        }

        private bool PrepareSimulation(Project project, ClearLevel clearLevel)
        {
            string simulationFolder = PathFinder.GetFolder(PathFinder.Folder.Simulation);

            foreach(string fileName in Directory.GetFiles(simulationFolder))
            {
                switch(clearLevel)
                {
                    case ClearLevel.All:
                        if(fileName != GetSimulatorPath()) { DeleteFile(fileName); }
                        break;

                    case ClearLevel.Minimum:
                        if (fileName != GetSimulatorPath() && Path.GetExtension(fileName) != ".vsf") { DeleteFile(fileName); }
                        break;

                    default:
                        if(Path.GetExtension(fileName) == ".set") { DeleteFile(fileName); }
                        break;
                }
            }

            if (!File.Exists(GetSimulatorPath()))
            {
                File.WriteAllBytes(GetSimulatorPath(), NMag.Core.Properties.Resources.nmag_DOS);
            }

            DataSetWriter dataSetWriter = new DataSetWriter();
            if (!dataSetWriter.Write(project, Path.Combine(simulationFolder, "nMAGset.set")).Succes) { return false; }

            if(clearLevel == ClearLevel.All)
            {
                RunoffSeriesWriter runoffSeriesWriter = new RunoffSeriesWriter();
                foreach(RunoffSeries series in project.Hydrology.RunoffSeries)
                {
                    if(!runoffSeriesWriter.Write(series, GetRunoffSeriesPath(series.Name.Replace(" ", ""))).Succes)
                    {
                        return false;
                    }
                }
            }

            return true;
        }
    }
}
