﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Controls.Primitives;
using System.Windows.Input;

namespace NMag.UI.CustomControls
{
    public class FilterComboBox : ComboBox
    {
        private List<string> _unfilteredItems = new List<string>();
        private string _filter = "";

        public FilterComboBox()
        {
            AddHandler(TextBoxBase.TextChangedEvent, new TextChangedEventHandler(OnTextChange));
        }

        public static readonly DependencyProperty AllItemsProperty = DependencyProperty.Register(
            "AllItems", 
            typeof(List<string>), 
            typeof(FilterComboBox),
            new FrameworkPropertyMetadata(new PropertyChangedCallback(AllItemsChanged)));

        public List<string> AllItems
        {
            get
            {
                return (GetValue(AllItemsProperty) as List<string>) ?? new List<string>();
            }
            set
            {
                SetValue(AllItemsProperty, value);
            }
        }

        private static void AllItemsChanged(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            List<string> allItems = (e.NewValue == null) ? new List<string>() : e.NewValue as List<string>;

            FilterComboBox comboBox = d as FilterComboBox;

            if (!comboBox._unfilteredItems.SequenceEqual(allItems))
            {
                comboBox._unfilteredItems = allItems;
                comboBox.ItemsSource = allItems;
            }
        }

        private string SelectedText
        {
            get
            {
                return (SelectedItem == null) ? "" : SelectedItem.ToString();
            }
        }

        protected override void OnKeyUp(KeyEventArgs e)
        {
            if(!char.IsLetterOrDigit(e.Key.ToString().First()))
            {
                return;
            }

            if ((e.Key == Key.Down || e.Key == Key.Up)
                && IsDropDownOpen == true
                && Keyboard.FocusedElement is TextBox)
            {
                Keyboard.Focus(this);
                return;
            }

            _filter = GetActualText();

            UpdateDropDownItems();
        }

        private void OnTextChange(object sender, TextChangedEventArgs e)
        {
            object selectedItem = AllItems
                .FirstOrDefault(i => i.ToString() == GetActualText());

            if(selectedItem != null)
            {
                SelectedItem = selectedItem;
                _filter = "";
            }
        }

        protected override void OnPreviewTextInput(TextCompositionEventArgs e)
        {
            base.OnPreviewTextInput(e);

            IsDropDownOpen = true;
        }

        protected override void OnDropDownOpened(EventArgs e)
        {
            base.OnDropDownOpened(e);

            UpdateDropDownItems();
        }

        protected override void OnSelectionChanged(SelectionChangedEventArgs e)
        {
            base.OnSelectionChanged(e);

            Text = SelectedText;
        }

        protected override void OnDropDownClosed(EventArgs e)
        {
            base.OnDropDownClosed(e);

            Text = SelectedText;

            _filter = "";
            ItemsSource = AllItems;
        }

        private string GetActualText()
        {
            return ((TextBox)Template.FindName("PART_EditableTextBox", this as FrameworkElement)).Text;
        }

        private void UpdateDropDownItems()
        {
            if (AllItems == null)
            {
                ItemsSource = null;
                return;
            }

            if(_filter == "")
            {
                if(SelectedText == "")
                {
                    ItemsSource = AllItems;
                }

                return;
            }

            ItemsSource = AllItems
                .Where(i => i.ToString().ToLower()
                    .Contains(_filter.ToLower()))
            .ToList();
        }
    }
}