﻿using LiveCharts;
using LiveCharts.Defaults;
using LiveCharts.Wpf;
using NMag.Core;
using NMag.Core.Utilization;
using System;
using System.Collections.Generic;
using System.Linq;

namespace NMag.UI.CustomControls
{
    public class SeriesData
    {
        private Curve _curve;

        public SeriesData(Curve curve, SeriesType type)
        {
            Type = type;
            Curve = curve;
        }

        public enum SeriesType
        {
            Line,
            Column
        }

        public Curve Curve
        {
            get
            {
                return _curve;
            }
            set
            {
                _curve = value;
            }
        }
        public SeriesType Type { get; set; }

        public bool IsTransposed { get; set; } = false;

        public bool IsDateMode { get; set; } = false;

        public Func<double, string> XFormatter
        {
            get
            {
                if(IsDateMode)
                {
                    switch(Curve.Axes[0].TimeScale)
                    {
                        case TimeScale.Hours:
                            return hour => new TimeSpan((int)hour, 0, 0).ToString("HH");;

                        case TimeScale.Days:
                            return day => DateHelper.DayToDate(2000, (int)day).ToString("MMM-dd");

                        case TimeScale.Months:
                            return m => new DateTime(2000, (int)Math.Min(12, Math.Max(1, m)), 1).ToString("MMM"); 

                        case TimeScale.Dates:
                            return d => d <= 0 ?
                                "" :
                                DateHelper.DayToDate(Curve.FirstYear, (int)d).ToString("yyyy-MM-dd");

                        default:
                            return v => v.ToString("G");
                    }
                }

                return v => v.ToString("G");
            }
        }

        public Func<double, string> YFormatter
        {
            get
            {
                return v => v.ToString("G");
            }
        }

        public Series GetSeries()
        {
            if((Curve?.Points.Count ?? 0) == 0)
            {
                return new LineSeries();
            }

            switch(Type)
            {
                case SeriesType.Line:
                    return CreateLineSeries();

                case SeriesType.Column:
                    return CreateColumnSeries();

                default:
                    return new LineSeries();
            }
        }

        private LineSeries CreateLineSeries()
        {
            ChartValues<ObservablePoint> values = new ChartValues<ObservablePoint>();

            List<Point> points = Curve.Type == CurveType.Hydro ?
                    new HydroCurve(_curve.Points.Select(p => p.Clone()).ToList(), _curve.FirstYear).GetAverage().Points :
                    _curve.Points;

            if (IsTransposed)
            {
                values.AddRange(points.Select(p => new ObservablePoint(p.Y, p.X)));
            }
            else
            {
                values.AddRange(points.Select(p => new ObservablePoint(p.X, p.Y)));
            }
            
            return new LineSeries
            {
                Values = values,
                PointGeometry = DefaultGeometries.None,
                LineSmoothness = 0
            };
        }

        private ColumnSeries CreateColumnSeries()
        {
            ChartValues<ObservablePoint> values = new ChartValues<ObservablePoint>();

            List<Point> points = Curve.Type == CurveType.Hydro ?
                    new HydroCurve(_curve.Clone().Points, _curve.FirstYear).GetAverage().Points :
                    _curve.Points;

            foreach (Point point in points)
            {
                values.Add(new ObservablePoint(point.X, point.Y));
            }

            return new ColumnSeries { Values = values};
        }
    }
}
