﻿using NMag.Core.Utilization;
using System.Windows.Controls;
using System.Windows.Input;
using System.Windows;
using NMag.UI.Dialogs;
using System;

namespace NMag.UI.CustomControls
{
    public class NumberOnlyTextBox : TextBox
    {
        public static readonly DependencyProperty ValueProperty = DependencyProperty.Register(
                "Value",
                typeof(double),
                typeof(NumberOnlyTextBox),
                new FrameworkPropertyMetadata(new PropertyChangedCallback(ValueChanged)));

        public double Value
        {
            get
            {
                return (double)GetValue(ValueProperty);
            }
            set
            {
                SetValue(ValueProperty, value);
            }
        }

        public static readonly DependencyProperty PrecisionProperty = DependencyProperty.Register(
                "Precision",
                typeof(int),
                typeof(NumberOnlyTextBox));

        public int Precision
        {
            get
            {
                return (int)GetValue(PrecisionProperty);
            }
            set
            {
                SetValue(PrecisionProperty, value);
            }
        }

        public static readonly DependencyProperty ShowDatesProperty = DependencyProperty.Register(
            "ShowDates",
            typeof(bool),
            typeof(NumberOnlyTextBox));

        public bool ShowDates
        {
            get
            {
                return (bool)GetValue(ShowDatesProperty);
            }
            set { }
        }

        public static readonly DependencyProperty FirstYearProperty = DependencyProperty.Register(
            "FirstYear",
            typeof(int),
            typeof(NumberOnlyTextBox));

        public int FirstYear
        {
            get
            {
                return (int)GetValue(FirstYearProperty);
            }
            set { }
        }

        protected override void OnLostKeyboardFocus(KeyboardFocusChangedEventArgs e)
        {
            FormatText();

            base.OnLostKeyboardFocus(e);
        }

        protected override void OnPreviewTextInput(TextCompositionEventArgs e)
        {
            if(Precision == 0 && e.Text == ".")
            {
                e.Handled = true;
                return;
            }

            if(SelectedText == Text && (e.Text.IsValidNumber() || e.Text == "."))
            {
                Text = e.Text;
                e.Handled = true;
                CaretIndex = Text.Length;
                return;
            }

            int caretIndex = CaretIndex;

            int dotIndex = Text.IndexOf('.');

            if(e.Text == ".")
            {
                if (dotIndex > 0 && caretIndex > dotIndex)
                {
                    caretIndex--;
                }

                Text = Text.Replace(".", "");

                CaretIndex = caretIndex;
            }

            e.Handled = !Util.IsValidNumber(Text.Insert(CaretIndex, e.Text));
        }

        protected override void OnTextChanged(TextChangedEventArgs e)
        {
            base.OnTextChanged(e);

            if (!IsFocused)
            {
                FormatText();
            }
        }

        protected override void OnMouseDown(MouseButtonEventArgs e)
        {
            base.OnMouseDown(e);

            if (!ShowDates)
            {
                return;
            }

            DateInputDialog dialog = new DateInputDialog(FirstYear) { DayNumber = (int)Value };
            dialog.ShowDialog();

            if (dialog.DialogResult == true)
            {
                Value = dialog.DayNumber;
                FormatText();
            }
        }

        protected override void OnInitialized(EventArgs e)
        {
            base.OnInitialized(e);

            FormatText();
        }

        private static void ValueChanged(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            if (d is NumberOnlyTextBox textBox)
            {
                textBox.Value = (double)e.NewValue;

                textBox.Text = textBox.Value.ToString();
            }
        }

        private void FormatText()
        {
            string formattedText;

            if(ShowDates)
            {
                if(FirstYear == 0)
                {
                    formattedText = DateHelper.DayToDate(1999, (int)Value).ToString("MMM-dd");
                }
                else
                {
                    formattedText = DateHelper.DayToDate(FirstYear, (int)Value).ToString("yyyy-MM-dd");
                }

                if(Text != formattedText)
                {
                    Text = formattedText;
                }
            }
            else
            {
                double.TryParse(Text, out double value);
                Value = value;

                formattedText = value.Format(Precision);

                if (Text != formattedText)
                {
                    Text = formattedText;
                }
            }
        }
    }
}
