﻿using NMag.Core.FlowDiagram;
using NMag.Core;
using NMag.Core.Utilization;
using System.Collections.Generic;
using System.Linq;
using System.Windows.Input;
using System.Windows.Media;

using Point = System.Windows.Point;
using System.Windows.Shapes;

namespace NMag.UI.FlowDiagram
{
    public class ModuleControlFactory
    {
        private const int COLUMN_WIDTH = 400;
        private const int LEVEL_HEIGHT = 60;
        private const int NUMBER_SPACING = 20;

        private const int MODULE_WIDTH = 300;
        private const int MODULE_HEIGHT = 35;

        private ShapeFactory _shapeFactory;
        private LabelFactory _labelFactory;

        private List<int> _rowHeights = new List<int>();

        public ModuleControlFactory(MouseButtonEventHandler onModuleClickHandler)
        {
            _shapeFactory = new ShapeFactory(onModuleClickHandler);
            _labelFactory = new LabelFactory(onModuleClickHandler);
        }

        public IEnumerable<ModuleControl> CreateModuleControls(List<Module> modules)
        {
            NMag.Core.FlowDiagram.FlowDiagram diagram = new FlowDiagramFactory().CreateFlowDiagram(modules);

            _rowHeights = new List<int>();

            for (int r = 0; r <= diagram.HighestRow; r++)
            {
                if(diagram.HigherRows.Contains(r))
                {
                    _rowHeights.Add(2);
                }
                else
                {
                    _rowHeights.Add(1);
                }
            }

            int levelStart = 0;

            for (int l = 0; l < diagram.FlowLevelHeights.Count; l++)
            {
                foreach (FlowChain chain in diagram.FlowChains.Where(c => c.Level == l))
                {
                    foreach (int moduleNumber in chain.ModuleNumbers)
                    {
                        Module module = modules.FirstOrDefault(m => m.Number == moduleNumber); 

                        yield return CreateModule(module, new Point(chain.Column * COLUMN_WIDTH, GetRowPosition(levelStart + chain.ModuleNumbers.IndexOf(moduleNumber))), diagram.Connections.Where(c => c.ModuleNumber == moduleNumber).ToList());
                    }
                }

                levelStart += diagram.FlowLevelHeights[l];
            }
        }

        private double GetRowCoordinate(double row, double offset = 0)
        {
            return (_rowHeights.GetRange(0, (int)row).Sum()) + offset;
        }

        private double GetRowPosition(double row, double offset = 0)
        {
            int correction = 0;

            if(_rowHeights[(int)row] == 2 && offset > 0)
            {
                correction = 1;
            }

            return (GetRowCoordinate(row, offset) + correction) * -LEVEL_HEIGHT;
        }

        private ModuleControl CreateModule(Module module, Point position, List<FlowConnection> connections)
        {
            if(module == null || module.Units.Count == 0) { return null; }

            List<ShapeBase> shapes = new List<ShapeBase>();
            List<Label> labels = new List<Label>();
            List<Line> lines = new List<Line>();

            if (module.ModuleType == ModuleType.Combined)
            {
                Unit reservoir = UnitHelper.GetReservoirs(module.Units).FirstOrDefault();
                Unit powerPlant = UnitHelper.GetPowerPlants(module.Units).FirstOrDefault();

                shapes.Add(_shapeFactory.CreateShape(reservoir.ID,
                    ShapeFactory.UnitTypeToShapeType(reservoir.UnitType),
                    position.Add(new Point(0, -(LEVEL_HEIGHT + MODULE_HEIGHT) / 2)),
                    MODULE_WIDTH, MODULE_HEIGHT));

                shapes.Add(_shapeFactory.CreateShape(powerPlant.ID,
                    ShapeFactory.UnitTypeToShapeType(powerPlant.UnitType),
                    position.Add(new Point(0, -(LEVEL_HEIGHT - MODULE_HEIGHT) / 2)),
                    MODULE_WIDTH, MODULE_HEIGHT));

                shapes.Add(_shapeFactory.CreateShape(reservoir.ID,
                    ShapeFactory.ShapeType.NumberShape,
                    position.Add(new Point(-MODULE_WIDTH / 2 - NUMBER_SPACING, - LEVEL_HEIGHT / 2)),
                    MODULE_WIDTH, MODULE_HEIGHT));

                labels.Add(_labelFactory.CreateTextBlock(reservoir.ID, reservoir.Name, shapes[0].Position.Add(new Point(0, 0)),
                    MODULE_WIDTH, MODULE_HEIGHT));
                labels.Add(_labelFactory.CreateTextBlock(powerPlant.ID, powerPlant.Name, shapes[1].Position.Add(new Point(0, 0)),
                    MODULE_WIDTH, MODULE_HEIGHT));

                labels.Add(_labelFactory.CreateTextBlock(reservoir.ID, module.Number.ToString(), shapes[2].Position,
                    MODULE_WIDTH, MODULE_HEIGHT));
            }
            else
            {
                Unit unit = module.Units.First();

                shapes.Add(_shapeFactory.CreateShape(unit.ID, 
                    ShapeFactory.UnitTypeToShapeType(unit.UnitType), 
                    position,
                    MODULE_WIDTH, MODULE_HEIGHT));

                shapes.Add(_shapeFactory.CreateShape(unit.ID, 
                    ShapeFactory.ShapeType.NumberShape, 
                    position.Add(new Point(-MODULE_WIDTH / 2 - NUMBER_SPACING, 0)),
                    MODULE_WIDTH, MODULE_HEIGHT));

                labels.Add(_labelFactory.CreateTextBlock(unit.ID, module.Name, shapes[0].Position,
                    MODULE_WIDTH, MODULE_HEIGHT));
                labels.Add(_labelFactory.CreateTextBlock(unit.ID, module.Number.ToString(), shapes[1].Position,
                    MODULE_WIDTH, MODULE_HEIGHT));
            }

            foreach(FlowConnection connection in connections)
            {
                lines.Add(new Line()
                {
                    Stroke = Brushes.Black,
                    StrokeThickness = 2,
                    StrokeDashArray = GetStrokeDashArray(connection.FlowConnectionType),
                    X1 = connection.Path.First().X * COLUMN_WIDTH,
                    Y1 = GetRowPosition(connection.Path.First().Y),
                    X2 = connection.Path.First().X * COLUMN_WIDTH,
                    Y2 = GetRowPosition(connection.Path.First().Y, -0.5),
                });

                if (connection.Path.Count > 2)
                {
                    lines.Add(new Line()
                    {
                        Stroke = Brushes.Black,
                        StrokeThickness = 2,
                        StrokeDashArray = GetStrokeDashArray(connection.FlowConnectionType),
                        X1 = connection.Path[0].X * COLUMN_WIDTH,
                        Y1 = GetRowPosition(connection.Path[0].Y, -0.5),
                        X2 = (connection.Path[1].X - 0.5) * COLUMN_WIDTH,
                        Y2 = GetRowPosition(connection.Path[1].Y, -0.5)
                    });

                    lines.Add(new Line()
                    {
                        Stroke = Brushes.Black,
                        StrokeThickness = 2,
                        StrokeDashArray = GetStrokeDashArray(connection.FlowConnectionType),
                        X1 = (connection.Path[1].X - 0.5) * COLUMN_WIDTH,
                        Y1 = GetRowPosition(connection.Path[1].Y, -0.5),
                        X2 = (connection.Path[2].X - 0.5) * COLUMN_WIDTH,
                        Y2 = GetRowPosition(connection.Path[2].Y, 0.5)
                    });

                    lines.Add(new Line()
                    {
                        Stroke = Brushes.Black,
                        StrokeThickness = 2,
                        StrokeDashArray = GetStrokeDashArray(connection.FlowConnectionType),
                        X1 = (connection.Path[2].X - 0.5) * COLUMN_WIDTH,
                        Y1 = GetRowPosition(connection.Path[2].Y, 0.5),
                        X2 = connection.Path[3].X * COLUMN_WIDTH,
                        Y2 = GetRowPosition(connection.Path[3].Y, 0.5)
                    });
                }

                lines.Add(new Line()
                {
                    Stroke = Brushes.Black,
                    StrokeThickness = 2,
                    StrokeDashArray = GetStrokeDashArray(connection.FlowConnectionType),
                    X1 = connection.Path.Last().X * COLUMN_WIDTH,
                    Y1 = GetRowPosition(connection.Path.Last().Y, 0.5),
                    X2 = connection.Path.Last().X * COLUMN_WIDTH,
                    Y2 = GetRowPosition(connection.Path.Last().Y)
                });
            }

            return new ModuleControl(module.Name, new Point(), shapes, labels, lines);
        }

        private DoubleCollection GetStrokeDashArray(FlowConnectionType type)
        {
            switch(type)
            {
                case FlowConnectionType.Bypass:
                    return new DoubleCollection(new double[] { 4, 8 });

                case FlowConnectionType.Spill:
                    return new DoubleCollection(new double[] { 1, 2 });

                default:
                    return null;
            }
        }
    }
}
