﻿using NMag.Core;
using NMag.UI.Dialogs;
using NMag.UI.Miscellaneous;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Windows;
using System.Windows.Input;
using System;
using System.Windows.Forms;
using System.ComponentModel;
using NMag.Core.Simulation;

namespace NMag.UI.Views
{
    public class ApplicationViewModel : WindowViewModel
    {
        private MainViewModel _currentPageViewModel;

        private Project _project;
        private SimulationResult _simulationResult;

        public ApplicationViewModel(EventHandler requestCloseHandler, string projectPath)
            : base(requestCloseHandler)
        {
            CreateViewModels();

            Update();
        }

        public string Version
        {
            get { return "Version: beta " + FileVersionInfo.GetVersionInfo(Assembly.GetExecutingAssembly().Location).FileVersion; }
        }

        public string ProjectName
        {
            get { return _project?.Name ?? ""; }
        }

        public Project Project
        {
            get { return _project; }
            set
            {
                _project?.Save(true);
                _project = value;
                _project.Save(true);

                CreateViewModels();
                Update();
            }
        }
        public bool IsNoProjectOpen
        {
            get { return Project == null; }
        }

        public SimulationResult SimulationResult
        {
            get { return _simulationResult; }
            set
            {
                _simulationResult = value;
                ViewModels.RemoveAll(v => v.ViewModelGroup == ViewModelGroup.Results);
                ViewModels.Add(new WaterSystemViewModel(this, ViewModelGroup.Results));
                ViewModels.Add(new ResultsViewModel(this));

                Update();
            }
        }

        public List<MainViewModel> ViewModels { get; set; }
        public List<MainViewModel> SetupViewModels
        {
            get { return ViewModels.Where(v => v.ViewModelGroup == ViewModelGroup.Setup).ToList(); }
        }
        public List<MainViewModel> SimulationViewModels
        {
            get { return ViewModels.Where(v => v.ViewModelGroup == ViewModelGroup.Simulation).ToList(); }
        }
        public List<MainViewModel> ResultsViewModels
        {
            get { return ViewModels.Where(v => v.ViewModelGroup == ViewModelGroup.Results).ToList(); }
        }
        public MainViewModel CurrentViewModel
        {
            get { return _currentPageViewModel; }
            private set
            {
                _currentPageViewModel = value;
                OnPropertyChanged("CurrentViewModel");
            }
        }

        public ToolbarViewModel ToolbarViewModel
        {
            get { return CurrentViewModel?.Toolbar; }
        }

        public ICommand ChangeViewCommand
        {
            get
            {
                return new RelayCommand(args => ChangeViewModel((MainViewModel)args),
                    args => args is MainViewModel);
            }
        }

        public ICommand CreateProjectCommand
        {
            get { return new RelayCommand(args => CreateNewProject()); }
        }

        public ICommand ImportDataSetCommand
        {
            get { return new RelayCommand(args => ImportDataSet()); }
        }
        public ICommand ExportDataSetCommand
        {
            get
            {
                return new RelayCommand(args => ExportDataSet(),
                    args => Project != null);
            }
        }

        public ICommand ShowAboutCommand
        {
            get { return new RelayCommand(args => new AboutDialog().ShowDialog()); }
        }

        public override void Update()
        {
            foreach(MainViewModel viewModel in ViewModels)
            {
                viewModel.SaveData();
                viewModel.Update();    
            }

            OnPropertyChanged("ProjectName");
            OnPropertyChanged("IsNoProjectOpen");
        }

        internal void OnWindowClosing(object sender, CancelEventArgs e)
        {
            if(Project != null)
            {
                MessageBoxResult result = System.Windows.MessageBox.Show("Do you want to save the current project?", "Closing Program", MessageBoxButton.YesNoCancel, MessageBoxImage.Question);

                switch(result)
                {
                    case MessageBoxResult.Cancel:
                        e.Cancel = true;
                        return;
                    case MessageBoxResult.Yes:
                        Project.Save(true);
                        break;
                    default:
                        break;
                }
            }

        }

        private void ChangeViewModel(MainViewModel viewModel)
        {
            if(viewModel == null) { return; }

            CurrentViewModel?.SaveData();

            viewModel.Update();
            CurrentViewModel = viewModel;
            CurrentViewModel.GetData();

            OnPropertyChanged("ToolbarViewModel");
            OnPropertyChanged("SetupViewModels");
            OnPropertyChanged("SimulationViewModels");
            OnPropertyChanged("ResultsViewModels");
        }

        private bool CanCloseProject()
        {
            if (Project != null)
            {
                if (System.Windows.MessageBox.Show("Are you sure you want to create a new project?\nThis project will be saved automatically.", "Create New Project", MessageBoxButton.YesNo, MessageBoxImage.Question) == MessageBoxResult.No)
                {
                    return false;
                }

                Project.Save(true);
            }

            return true;
        }

        private void CreateNewProject()
        {
            if(CanCloseProject() == false) { return; }

            TextInputDialog dialog = new TextInputDialog("Project Name", "What is the name of the new project?", "No name given!");
            if(dialog.ShowDialog() == true && (dialog.Result ?? "") != "")
            {
                if(File.Exists(PathFinder.GetFullFileName(PathFinder.FileType.Project, dialog.Result)))
                {
                    if (System.Windows.MessageBox.Show("A project with this name already exists.\nDo you want to overwrite the existing project?.", "Overwrite Existing Project", MessageBoxButton.YesNo, MessageBoxImage.Question) == MessageBoxResult.No)
                    {
                        return;
                    }
                }

                Result result = ProjectHelper.CreateProject(dialog.Result, out Project project, true);

                if (result.Succes) { Project = project; }
            }
        }

        private void ImportDataSet()
        {
            if (CanCloseProject() == false) { return; }

            OpenFileDialog fileDialog = new OpenFileDialog()
            {
                Title = "Select a NMag Data Set",
                Filter = "NMag Data Sets (*.set)|*.set"
            };

            if(fileDialog.ShowDialog() == DialogResult.OK)
            {
                Result result = new DataSetReader().Read(fileDialog.FileName, out Project project);

                if(!result.Succes)
                {
                    System.Windows.MessageBox.Show(result.Message, "Data Set Read Error", MessageBoxButton.OK, MessageBoxImage.Error);
                    return;
                }

                if(File.Exists(PathFinder.GetFullFileName(PathFinder.FileType.Project, project.Name)))
                {
                    if (System.Windows.MessageBox.Show($"A project named {project.Name} already exists.\nDo you want to overwrite the existing project?.", "Overwrite Existing Project", MessageBoxButton.YesNo, MessageBoxImage.Question) == MessageBoxResult.No)
                    {
                        return;
                    }
                }

                Project = project;
            }
        }

        private void ExportDataSet()
        {
            string fileName = PathFinder.GetFullFileName(PathFinder.FileType.DataSet, ProjectName, ProjectName);

            Result result = new DataSetWriter().Write(Project, fileName);

            if (result.Succes)
            {
                Process.Start(fileName);
            }
            else
            {
                System.Windows.MessageBox.Show(result.Message, "Error Writing Data Set", MessageBoxButton.OK, MessageBoxImage.Error);
            }
        }

        private void CreateViewModels()
        {
            ViewModels = new List<MainViewModel>()
            {
                new WaterSystemViewModel(this, ViewModelGroup.Setup),
                new HydrologyViewModel(this),
                new EnergyMarketViewModel(this),
                new SimulationViewModel(this),
                new WaterSystemViewModel(this, ViewModelGroup.Results),
                new ResultsViewModel(this)
            };

            OnPropertyChanged("ViewModels");
            ChangeViewModel(ViewModels.First());
        }
    }
}
