﻿using NMag.Core;
using System.Collections.Generic;
using System.Linq;

namespace NMag.UI.Views
{
    public class ModuleSetupViewModel : ModuleViewModel
    {
        private ViewModel _routingViewModel;
        private ViewModel _hydrologySelectionViewModel;
        private ViewModel _restrictionViewModel;
        private EnMagStrategyViewModel _enMagStrategyViewModel;
        private NMagStrategyViewModel _nMagStrategyViewModel;
        private ViewModel _currentDataViewModel;

        public ModuleSetupViewModel(WaterSystemViewModel waterSystem)
            : base(waterSystem)
        {
            CreateViewModels();
            GetData();
        }

        public ViewModel RoutingViewModel
        {
            get { return _routingViewModel; }
        }
        public ViewModel HydrologySelectionViewModel
        {
            get { return _hydrologySelectionViewModel; }
        }
        public ViewModel RestrictionViewModel
        {
            get { return _restrictionViewModel; }
        }

        public bool UseOperationalStrategy
        {
            get { return SelectedModule?.UseOperationalStrategy ?? false; }
            set
            {
                if (SelectedModule == null) { return; }
                
                SelectedModule.UseOperationalStrategy = value;
                Update();
            }
        }
        public bool CanUseEnMagMode
        {
            get { return WaterSystem.Application.Project == null ? false : WaterSystem.Application.Project?.Modules.Count(m => m.ModuleType == ModuleType.Reservoir || m.ModuleType == ModuleType.Combined) == 1; }
        }
        public bool UseEnMagStrategy
        {
            get { return CanUseEnMagMode && WaterSystem.Application.Project.EnMagMode; }
            set
            {
                if(WaterSystem.Application.Project == null) { return; }

                WaterSystem.Application.Project.EnMagMode = value && CanUseEnMagMode;
                Update();
            }
        }
        public OperationalStrategyViewModel StrategyViewModel
        {
            get
            {
                if(WaterSystem.Application.Project == null) { return null; }

                if(WaterSystem.Application.Project.EnMagMode && CanUseEnMagMode)
                {
                    return _enMagStrategyViewModel;
                }

                return _nMagStrategyViewModel;
            }
        }

        public ViewModel DataViewModel
        {
            get
            {
                if (_currentDataViewModel != null) { _currentDataViewModel.SaveData(); }

                switch (SelectedUnitType)
                {
                    case "0":
                        return _currentDataViewModel = new ReservoirViewModel(SelectedUnit as ReservoirUnit);

                    case "1":
                        return _currentDataViewModel = new PowerPlantViewModel(SelectedUnit as PowerPlantUnit);

                    case "2":
                        return _currentDataViewModel = new TransferViewModel(SelectedUnit as TransferUnit);

                    case "3":
                        return _currentDataViewModel = new ControlPointViewModel(SelectedUnit as ControlPointUnit);

                    default: return null;
                }
            }
        }

        public override void GetData()
        {
            RoutingViewModel.GetData();
            HydrologySelectionViewModel?.GetData();
            RestrictionViewModel?.GetData();

            _nMagStrategyViewModel?.GetData();
            _enMagStrategyViewModel?.GetData();

            DataViewModel?.GetData();
        }

        public override void SaveData()
        {
            RoutingViewModel?.SaveData();
            HydrologySelectionViewModel?.SaveData();
            RestrictionViewModel?.SaveData();

            _nMagStrategyViewModel?.SaveData();
            _enMagStrategyViewModel?.SaveData();

            DataViewModel?.SaveData();
        }

        public override void Update()
        {
            OnPropertyChanged("ModuleTypes");
            OnPropertyChanged("SelectedModuleType");

            OnPropertyChanged("SelectedName");
            OnPropertyChanged("IsModuleSelected");
            OnPropertyChanged("IsReservoirModule");

            OnPropertyChanged("SelectedUnit");
            OnPropertyChanged("SelectedUnitType");

            OnPropertyChanged("HasReservoirUnit");
            OnPropertyChanged("HasPowerPlantUnit");
            OnPropertyChanged("HasTransferUnit");
            OnPropertyChanged("HasControlPointUnit");

            OnPropertyChanged("UseOperationalStrategy");
            OnPropertyChanged("CanUseEnMagMode");
            OnPropertyChanged("UseEnMagStrategy");

            _routingViewModel.Update();
            _hydrologySelectionViewModel.Update();
            _restrictionViewModel.Update();
            OnPropertyChanged("DataViewModel");

            _nMagStrategyViewModel.Update();
            _enMagStrategyViewModel.Update();
            OnPropertyChanged("StrategyViewModel");
        }

        private void CreateViewModels()
        {
            _routingViewModel = new RoutingViewModel(this);
            _hydrologySelectionViewModel = new HydrologySelectionViewModel(this);
            _restrictionViewModel = new RestrictionViewModel(this);
            _enMagStrategyViewModel = new EnMagStrategyViewModel(this);
            _nMagStrategyViewModel = new NMagStrategyViewModel(this);
        }
    }
}
