﻿using NMag.UI.Miscellaneous;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Windows.Input;
using System.Windows.Media;
using System.Linq;
using NMag.UI.FlowDiagram;
using NMag.UI.Dialogs;
using NMag.Core;
using System.Windows;

namespace NMag.UI.Views
{
    public class WaterSystemViewModel : MainViewModel
    {
        private ModuleModel _moduleModel;
        private ModuleSetupViewModel _moduleViewModel;

        public WaterSystemViewModel(ApplicationViewModel application, ViewModelGroup group)
                :base(application, "Water System", group)
        {
            CanvasControls = new CanvasControls(this);
            _moduleModel = new ModuleModel(new MouseButtonEventHandler(OnModuleClick));
            _moduleViewModel = new ModuleSetupViewModel(this);

            GetData();
        }

        public override ToolbarViewModel Toolbar
        {
            get { return base.ViewModelGroup == ViewModelGroup.Setup ? new WaterSystemToolbarViewModel(this) : null; }
        }

        public ObservableCollection<object> ModuleData
        {
            get
            {
                List<object> objects = new List<object>();
                objects.AddRange(_moduleModel.GetConnections());
                objects.AddRange(_moduleModel.GetShapes());
                objects.AddRange(_moduleModel.GetTextBlocks());

                return new ObservableCollection<object>(objects);
            }
        }

        public CanvasControls CanvasControls { get; private set; }

        public MatrixTransform RenderTransform
        {
            get { return CanvasControls.RenderTransform; }
        }

        public ModuleSetupViewModel ModuleViewModel
        {
            get { return _moduleViewModel; }
            set { _moduleViewModel = value; }
        }

        public double MouseX
        {
            get { return CanvasControls.MouseX; }
            set { CanvasControls.MouseX = value; }
        }
        public double MouseY
        {
            get { return CanvasControls.MouseY; }
            set { CanvasControls.MouseY = value; }
        }

        public ICommand ZoomInCommand
        {
            get { return new RelayCommand(args => CanvasControls.OnZoom(args, 1)); }
        }
        public ICommand ZoomOutCommand
        {
            get { return new RelayCommand(args => CanvasControls.OnZoom(args, -1)); }
        }

        public ICommand MouseDownCommand
        {
            get { return new RelayCommand(args => CanvasControls.OnMouseDown((MouseEventArgs)args)); }
        }
        public ICommand MouseUpCommand
        {
            get { return new RelayCommand(args => CanvasControls.OnMouseUp((MouseEventArgs)args)); }
        }

        public ICommand MouseMoveCommand
        {
            get { return new RelayCommand(args => CanvasControls.OnMouseMove()); }
        }
        public ICommand ResizeCommand
        {
            get { return new RelayCommand(args => CanvasControls.OnChangeCanvasSize(args)); }
        }

        public void OnModuleClick(object sender, MouseButtonEventArgs e)
        {
            string unitID = "";

            if (sender.GetType().BaseType == typeof(ShapeBase))
            {
                unitID = ((ShapeBase)sender).UnitID.ToString();
            }

            if (sender.GetType() == typeof(Label))
            {
                unitID = ((Label)sender).UnitID.ToString();
            }

            if (unitID != "")
            {
                CanvasControls.CancelMove = true;

                SaveData();
                ModuleViewModel.SelectedModule = Application.Project.Modules
                    .FirstOrDefault(m => m.Units
                        .Select(u => u.ID.ToString())
                        .Contains(unitID));

                ModuleViewModel.SelectedUnit = ModuleViewModel.SelectedModule.Units.FirstOrDefault(u => u.ID.ToString() == unitID);
                Update();
            }
        }

        public void UpdateCanvas()
        {
            OnPropertyChanged("RenderTransform");
        }

        public void AddModule()
        {
            if(Application.Project == null) { return; }

            ModuleDialog moduleDialog = new ModuleDialog();
            if(moduleDialog.ShowDialog() == true)
            {
                List<string> unitNames = Application.Project.Modules.SelectMany(m => m.Units.Select(u => u.Name)).ToList();

                if(unitNames.Contains(moduleDialog.Names[0]))
                {
                    MessageBox.Show($"{moduleDialog.Names[0]} already exists.", "Duplicate Name", MessageBoxButton.OK, MessageBoxImage.Error);
                    return;
                }

                if (moduleDialog.Names[1] != "" && unitNames.Contains(moduleDialog.Names[1]))
                {
                    MessageBox.Show($"{moduleDialog.Names[1]} already exists.", "Duplicate Name", MessageBoxButton.OK, MessageBoxImage.Error);
                    return;
                }

                int moduleNumber = Application.Project.Modules.Count == 0 ?
                    1 :
                    Application.Project.Modules.Max(m => m.Number) + 1;

                Module module = ModuleHelper.CreateModule((ModuleType)moduleDialog.Type, moduleNumber, moduleDialog.Names[0], moduleDialog.Names[1]);

                if(module != null)
                {
                    Application.Project.Modules.Add(module);
                    Application.Update();
                }
            }
        }

        public void RenameModule()
        {
            if (Application.Project == null) { return; }

            Module module = ModuleViewModel.SelectedModule;

            string name1 = module.Name;
            string name2 = "";

            if(module.ModuleType == ModuleType.Combined)
            {
                name1 = UnitHelper.GetReservoirs(module.Units).First().Name;
                name2 = UnitHelper.GetPowerPlants(module.Units).First().Name;
            }

            ModuleDialog moduleDialog = new ModuleDialog(name1, name2);
            if (moduleDialog.ShowDialog() == true)
            {
                List<string> unitNames = Application.Project.Modules.SelectMany(m => m.Units.Select(u => u.Name)).Where(n => n != name1 && n != name2).ToList();

                if (unitNames.Contains(moduleDialog.Names[0]))
                {
                    MessageBox.Show($"{moduleDialog.Names[0]} already exists.", "Duplicate Name", MessageBoxButton.OK, MessageBoxImage.Error);
                    return;
                }

                if (moduleDialog.Names[1] != "" && unitNames.Contains(moduleDialog.Names[1]))
                {
                    MessageBox.Show($"{moduleDialog.Names[1]} already exists.", "Duplicate Name", MessageBoxButton.OK, MessageBoxImage.Error);
                    return;
                }

                if (module.ModuleType == ModuleType.Combined)
                {
                    UnitHelper.GetReservoirs(module.Units).First().Name = moduleDialog.Names[0];
                    UnitHelper.GetPowerPlants(module.Units).First().Name = moduleDialog.Names[1];
                }
                else
                {
                    module.Units.First().Name = moduleDialog.Names[0];
                }

                Application.Update();
            }
        }

        public void DeleteModule()
        {
            if (Application.Project == null || !ModuleViewModel.IsModuleSelected) { return; }

            Module module = ModuleViewModel.SelectedModule;

            if(MessageBox.Show($"Are you sure you want to remove module #{module.Number} - {module.Name}?\nThis action cannot be undone.", "Delete Module", MessageBoxButton.YesNo, MessageBoxImage.Warning) == MessageBoxResult.Yes)
            {
                Application.Project.Modules.Remove(module);

                foreach(Module connectedModule in Application.Project.Modules.Where(m => m.RoutingNumbers.Contains(module.Number)))
                {
                    for (int i = 0; i < 3; i++)
                    {
                        if(connectedModule.ReleaseModuleNumber == module.Number) { connectedModule.ReleaseModuleNumber = module.ReleaseModuleNumber; }
                        if (connectedModule.BypassModuleNumber == module.Number) { connectedModule.BypassModuleNumber = module.ReleaseModuleNumber; }
                        if (connectedModule.SpillModuleNumber == module.Number) { connectedModule.SpillModuleNumber = module.ReleaseModuleNumber; }
                    }
                }
                
                Application.Update();
            }
        }

        public override void GetData()
        {
            ModuleViewModel.GetData();
        }

        public override void SaveData()
        {
            ModuleViewModel.SaveData();
        }

        public override void Update()
        {
            UpdateToolbar();

            switch(ViewModelGroup)
            {
                case ViewModelGroup.Setup:
                    if (Application.Project != null) { _moduleModel.UpdateModules(Application.Project.Modules); }
                    break;
                case ViewModelGroup.Results:
                    if (Application.SimulationResult != null) { _moduleModel.UpdateModules(Application.SimulationResult.ModuleResults.Select(m => m.Module).ToList()); }
                    break;
                default:
                    break;
            }

            OnPropertyChanged("ModuleData");

            _moduleViewModel.Update();
        }
    }
}
